/**
 * JdbcTool: Command line pain relief for JDBC databases.
 * Copyright (C) 2007, Quuxo Software.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

package com.quuxo.jdbctool;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Statement;
import java.util.LinkedList;

import gnu.getopt.Getopt;


/**
 * Command line program to dump a database as a series of SQL statements.
 *
 * @author <a href="mailto:michael@quuxo.com">Michael Gratton</a>
 */
public class JdbcDump {


    protected String[] driverNames = new String[] { "org.hsqldb.jdbcDriver" };
    protected Connection connection;
    protected String eol = System.getProperty("line.separator");
    protected String operation;

    protected boolean addDelete = true;
    
    
    public JdbcDump(String url, String username, String password)
        throws SQLException {
        loadDrivers();
        this.connection = DriverManager.getConnection(url, username, password);
    }

    protected void loadDrivers() {
        for (String name: this.driverNames) {
            try {
                Class.forName(name);
            } catch (Exception e) {
                // XXX log me
            }
        }
    }

    public void close()
        throws SQLException {
        this.connection.close();
    }

    public void dump(Writer sql)
        throws SQLException, IOException {
        this.operation = "Retreiving table names";
        DatabaseMetaData dbmeta = this.connection.getMetaData();
        ResultSet tables = dbmeta.getTables(null, null, null,
                                            new String[] { "TABLE" });
        LinkedList<String> tableNames = new LinkedList<String>();
        while (tables.next()) {
            tableNames.add(tables.getString(3));
            exhumeWarnings(tables);
        }
        tables.close();

        for (String table: tableNames) {
            dumpTable(table, sql);
        }
    }

    protected void dumpTable(String table, Writer sql)
        throws SQLException, IOException {
        this.operation = "Examining table " + table;

        sql.write(this.eol);
        sql.write(this.eol);

        Statement s = this.connection.createStatement();
        ResultSet results = s.executeQuery("SELECT * FROM " + table);
        ResultSetMetaData meta = results.getMetaData();
        int cols = meta.getColumnCount();

        String[] names = new String[cols];
        for (int i = 0; i < cols; i++) {
            names[i] = meta.getColumnName(i + 1);
        }

        if (this.addDelete) {
            sql.write("DELETE FROM ");
            sql.write(table);
            sql.write(";");
            sql.write(this.eol);
            sql.write(this.eol);
        }

        String[] values = new String[cols];
        int row = 0;
        while (results.next()) {
            this.operation = "Dumping table " + table + ", row " + row;
            for (int i = 0; i < cols; i++) {
                values[i] = results.getString(i + 1);
            }
            exhumeWarnings(results);
            row++;

            sql.write("INSERT INTO ");
            sql.write(table);
            sql.write(" (");
            sql.write(names[0]);
            for (int i = 1; i < cols; i++) {
                sql.write(", ");
                sql.write(names[i]);
            }
            sql.write(") VALUES (");
            sql.write(values[0]);
            for (int i = 1; i < cols; i++) {
                sql.write(", ");
                writeValue(values[i], sql);
            }
            sql.write(");");
            sql.write(this.eol);
        }
        s.close();
    }

    protected void writeValue(String value, Writer sql)
        throws IOException {
        if (value != null) {
            sql.write("'");
            for (int i = 0; i < value.length(); i++) {
                char c = value.charAt(i);
                sql.write(c);
                if (c == '\'') {
                    sql.write(c);
                }
            }
            sql.write("'");
        } else {
            sql.write("NULL");
        }
    }

    protected void exhumeWarnings(ResultSet results)
        throws SQLException {
        SQLWarning w = results.getWarnings();
        while (w != null) {
            printDumpWarning(w.toString());
            w = w.getNextWarning();
        }
        results.clearWarnings();
    }

    protected void printDumpWarning(String message) {
        System.err.println("Warning: " + this.operation + ": " + message);
    }

    protected void printDumpError(String message) {
        System.err.println("Error: " + this.operation + ": " + message);
    }

    protected static void printError(String message) {
        System.err.println(message);
    }

    protected static String promptUser(String message) {
        try {
            BufferedReader in =
                new BufferedReader(new InputStreamReader(System.in));
            System.err.println(message + ": ");
            return in.readLine();
        } catch (IOException ioe) {
            printError(ioe.getMessage());
            return null;
        }
    }

    public static void main(String[] argv) {
        String file = null;
        String password = null;
        String user = null;

        Getopt g = new Getopt("JdbcDump", argv, "f:p:Pu:");
        int c;
        while ((c = g.getopt()) != -1) {
            switch(c) {
            case 'f':
                file = g.getOptarg();
                break;
            case 'p':
                password = g.getOptarg();
                break;
            case 'P':
                password = promptUser("Enter password");
                break;
            case 'u':
                user = g.getOptarg();
                break;
            default:
                printError("JdbcDump: unknown option `" + c + "'");
                System.exit(-1);
            }
        }        

        int i = g.getOptind();
        if (i >= argv.length) {
            printError("No JDBC URL specified.");
            System.exit(-1);
        }
        String url = argv[i];

        BufferedWriter out = null;
        try {
            if (file != null) {
                out = new BufferedWriter(new FileWriter(file));
            } else {
                out = new BufferedWriter(new OutputStreamWriter(System.out));
            }
        } catch (IOException ioe) {
            printError("Unable to open SQL file: " + ioe);
            System.exit(-1);
        }


        JdbcDump jl = null;
        try {
            jl = new JdbcDump(url, user, password);
        } catch (SQLException sqle) {
            printError("Unable to connect to database: " + sqle);
            System.exit(-2);
        }

        try {
            jl.dump(out);
        } catch (Exception e) {
            jl.printDumpError(e.getMessage());
            e.printStackTrace();
            System.exit(-3);
        } finally {
            try {
                out.close();
            } catch (IOException ioe) {
                // oh well
            }
        }

        try {
            jl.close();
        } catch (SQLException sqle) {
            printError("Error closing connection: " + sqle);
            System.exit(-4);
        }

    }

}
