/**
 * JdbcTool: Command line pain relief for JDBC databases.
 * Copyright (C) 2007, Quuxo Software.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

package com.quuxo.jdbctool;

import java.io.BufferedReader;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Statement;
import java.util.ArrayList;

import gnu.getopt.Getopt;

import org.gnu.readline.Readline;
import org.gnu.readline.ReadlineLibrary;


/**
 * Command line program to execute SQL statements interactively.
 *
 * @author <a href="mailto:michael@quuxo.com">Michael Gratton</a>
 */
public class JdbcTool {


    protected String[] driverNames = new String[] { "org.hsqldb.jdbcDriver" };
    protected Connection connection;
    protected Statement statement;
    protected char eol = System.getProperty("line.separator").charAt(0);
    protected String prompt;
    protected String history;
    
    public JdbcTool(String url, String username, String password)
        throws SQLException {
        loadDrivers();
        this.connection = DriverManager.getConnection(url, username, password);
        setPrompt(url.startsWith("jdbc:")
                  ? url.substring("jdbc:".length()) : url);
        this.statement = this.connection.createStatement();
        this.history = System.getProperty("user.home") + "/.jdbctool_history";

        // readline init
        try {
            Readline.load(ReadlineLibrary.GnuReadline);
        }
        catch (UnsatisfiedLinkError ule) {
            System.err.println
                ("Java-Readline not found, using simple stdin.");
        }
        Readline.initReadline("JdbcTool");
        try {
            Readline.readHistoryFile(this.history);
        } catch (Exception e) {
            e.printStackTrace();
            // oh well
        }
    }

    protected void loadDrivers() {
        for (String name: this.driverNames) {
            try {
                Class.forName(name);
            } catch (Exception e) {
                // XXX log me
            }
        }
    }

    public void setPrompt(String prompt) {
        if (prompt == null) {
            prompt = "jdbctool";
        }
        this.prompt = prompt + "> ";
    }

    public void close()
        throws SQLException {
        this.connection.close();
        try {
            Readline.writeHistoryFile(this.history);
        } catch (Exception e) {
            // oh well
        }
        Readline.cleanup();
    }

    public void start()
        throws SQLException, IOException {
        // let's go!
        while (true) {
            try {
                String line = Readline.readline(this.prompt);
                if (line != null) {
                    execute(line);
                }
            } catch (EOFException e) {
                System.out.println();
                break;
            } catch (SQLException e) {
                printLineError(e.toString());
            }
        }
    }

    protected void execute(String line)
        throws SQLException {
        boolean hasResults = this.statement.execute(line);
        if (hasResults) {
            printResults();
        } else {
            System.out.println();
            System.out.println("Updated: " + this.statement.getUpdateCount());
            System.out.println();
        }
    }

    protected void printResults()
        throws SQLException {
        ResultSet results = this.statement.getResultSet();
        exhumeWarnings(results);

        ResultSetMetaData meta = results.getMetaData();
        int cols = meta.getColumnCount();

        String[] names = new String[cols];
        for (int i = 0; i < cols; i++) {
            names[i] = meta.getColumnName(i + 1);
        }

        int[] widths = new int[cols];
        int maxRows = 100;
        ArrayList<String[]> rows = new ArrayList<String[]>(maxRows);
        while (results != null) {
            // init widths
            for (int i = 0; i < cols; i++) {
                widths[i] = names[i].length();
            }

            // get values
            for (int row = 0; row < maxRows; row++) {
                if (!results.next()) {
                    results.close();
                    results = null;
                    break;
                }
                String[] values = new String[cols];
                for (int i = 0; i < cols; i++) {
                    values[i] = results.getString(i + 1);
                    if (values[i] == null) {
                        values[i] = "<NULL>";
                    }
                    widths[i] = Math.max(widths[i], values[i].length());
                }
                rows.add(values);
                exhumeWarnings(results);
            }

            // print it
            printLine(widths);
            printRow(names, widths);
            printLine(widths);
            for (String[] values: rows) {
                printRow(values, widths);
            }
            printLine(widths);
        }
        System.out.println();
    }

    protected void printRow(String[] values, int[] widths) {
        System.out.print("|");
        for (int i = 0; i < values.length; i++) {
            System.out.print(" ");
            System.out.print(values[i]);
            for (int w = values[i].length(); w < widths[i]; w++) {
                System.out.print(" ");
            }
            System.out.print(" |");
        }
        System.out.println();
    }

    protected void printLine(int[] widths) {
        System.out.print("-");
        for (int i = 0; i < widths.length; i++) {
            for (int w = -3; w < widths[i]; w++) {
                System.out.print("-");
            }
        }
        System.out.println();
    }

    protected void exhumeWarnings(ResultSet results)
        throws SQLException {
        SQLWarning w = results.getWarnings();
        while (w != null) {
            printLineWarning(w.toString());
            w = w.getNextWarning();
        }
        results.clearWarnings();
    }

    protected static String promptUser(String message) {
        try {
            BufferedReader in =
                new BufferedReader(new InputStreamReader(System.in));
            System.err.println(message + ": ");
            return in.readLine();
        } catch (IOException ioe) {
            printError(ioe.getMessage());
            return null;
        }
    }

    protected void printLineWarning(String message) {
        System.out.println("Warning: " + message);
    }

    protected void printLineError(String message) {
        System.out.println("Error: " + message);
    }

    protected static void printError(String message) {
        System.err.println(message);
    }

    public static void main(String[] argv) {
        String password = null;
        String user = null;

        Getopt g = new Getopt("JdbcTool", argv, "p:Pu:");
        int c;
        while ((c = g.getopt()) != -1) {
            switch(c) {
            case 'p':
                password = g.getOptarg();
                break;
            case 'P':
                password = promptUser("Enter password");
                break;
            case 'u':
                user = g.getOptarg();
                break;
            default:
                printError("JdbcTool: unknown option `" + c + "'");
                System.exit(-1);
            }
        }        

        int i = g.getOptind();
        if (i >= argv.length) {
            printError("No JDBC URL specified.");
            System.exit(-1);
        }
        String url = argv[i];
        JdbcTool jt = null;
        try {
            jt = new JdbcTool(url, user, password);
        } catch (SQLException sqle) {
            printError("Unable to connect to database: " + sqle);
            System.exit(-2);
        }

        try {
            jt.start();
        } catch (Exception e) {
            jt.printError(e.getMessage());
            System.exit(-3);
        }

        try {
            jt.close();
        } catch (SQLException sqle) {
            printError("Error closing connection: " + sqle);
            System.exit(-4);
        }

    }

}
